"""
.. module:: readspec
   :synopsis: Reads in a HASP spectrum from a FITS file.

.. moduleauthor:: Rob Swaters <rswaters@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
#--------------------
# External Imports
#--------------------
from astropy.io import fits
#--------------------
# Package Imports
#--------------------
from spec_plots.utils.specutils_hasp.haspspectrum import HASPSpectrum
from spec_plots import __version__

#--------------------

#--------------------
def readspec(input_file):
    """
    Reads in a HASP spectrum FITS file (_cspec.fits) and returns the
    wavelengths, fluxes, and flux uncertainties.

    :param input_file: Name of input FITS file.

    :type input_file: str

    :returns: HASPSpectrum -- The spectroscopic data (wavelength, flux, flux
        error, etc):
    """

    with fits.open(input_file) as hdulist:
        # Create an initially empty list that will contain each extension's
        # (association's) spectrum object.

        try:
            title = hdulist[0].header['TARGNAME']
        except KeyError:
            title = ""

        plot_info = {"title": title, "orig_file": input_file}

        # Data are in the first extension.
        data_table = hdulist[1].data

        return HASPSpectrum(
            nelem=len(data_table["WAVELENGTH"][0]),
            wavelengths=data_table["WAVELENGTH"][0],
            fluxes=data_table["FLUX"][0],
            fluxerrs=data_table["ERROR"][0],
            dqs=[0 if x > 0 else 1 for x in data_table["EFF_EXPTIME"][0]],
            plot_info=plot_info
            )

#--------------------



"""
.. module:: __init__

   :synopsis: This is the top-level of the package. Defines version of software.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

from __future__ import absolute_import

__version__ = '1.37.1'



"""
.. module:: stis1dspectrum
   :synopsis: Defines the class for STIS 1D spectra.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
#--------------------
# External Imports
#--------------------
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__

#--------------------

class STIS1DSpectrum():
    """
    Defines a STIS 1D spectrum (either "x1d" extracted or "sx1" summed
    extracted), including wavelegnth, flux, and flux errors.  A STIS 1D
    spectrum object consists of N associations.  If the file is an association,
    then N > 1, otherwise N = 1.  Each of these N associations can contain M
    orders.  If the association is an Echelle spectrum, then 24 < M < 70,
    depending on instrument configuration, otherwise M = 1.  Each of these M
    orders contain typical spectral data (wavelengths, fluxes, etc.), stored
    as STISOrderSpectrum objects.  The final data structure is then
    <STIS1DSpectrum>.associations[n].order[m].wavelengths (or .fluxes,
    .fluxerrs, etc.).

    :raises: ValueError
    """

    def __init__(self, association_spectra, orig_file=None):
        """
        Create a STIS1DSpectrum object out of a list of STISExposureSpectrum
        objects, which themselves are lists of STISOrderSpectrum objects.

        :param association_spectra: A list whose length is equal to the number
            of associations (length = "N" associations).  Each element in this
            list is a STISExposureSpectrum object, which itself is a list
            (length = "M" orders) of STISOrderSpectrum objects.

        :type association_spectra: list

        :param orig_file: Original FITS file read to create the spectrum
            (includes full path).

        :type orig_file: str
        """

        # Record the original file name along with the list of associations.
        self.orig_file = orig_file
        self.associations = association_spectra
#--------------------

#--------------------
class STISExposureSpectrum():
    """
    Defines a STIS exposure spectrum, which consists of "M" STISOrderSpectrum
    objects.
    """
    def __init__(self, order_spectra):
        """
        Create a STISExposureSpectrum object out of a list of STISOrderSpectrum
        objects.

        :param order_spectra: The STISOrderSpectrum objects to build the
            STISExposureSpectrum object out of.

        :type order_spectra: list

        :raises: ValueError
        """

        if len(order_spectra) > 0:
            self.orders = order_spectra
        else:
            raise ValueError("Must provide a list of at least one"
                             " STISOrderSpectrum object, input list is empty.")
#--------------------

#--------------------
class STISOrderSpectrum():
    """
    Defines a STIS order spectrum, including wavelength, flux, flux errors,
    and data quality flags, which are stored as numpy arrays.  A scalar int
    property provides the number of elements in this segment.
    """

    def __init__(self, nelem=None, wavelengths=None, fluxes=None, fluxerrs=None,
                 dqs=None):

        """
        Create a STISOrderSpectrum object, default to empty values.  Allows user
        to preallocate space if they desire by setting "nelem" but not providing
        lists/arrays on input right away.

        :param nelem: Number of elements for this segment's spectrum.

        :type nelem: int

        :param wavelengths: List of wavelengths in this segment's spectrum.

        :type wavelengths: list

        :param fluxes: List of fluxes in this segment's spectrum.

        :type fluxes: list

        :param fluxerrs: List of flux uncertainties in this segment's spectrum.

        :type fluxerrs: list

        :param dqs: List of data quality flags.

        :type dqs: list
        """

        # <DEVEL> Should it be required to have `nelem` > 0 *OR* specify
        # arrays on input?  Otherwise they are pre-allocated to empty
        # lists. </DEVEL>
        if nelem is not None:
            self.nelem = nelem
        else:
            self.nelem = 0

        if wavelengths is not None:
            self.wavelengths = numpy.asarray(wavelengths)
        else:
            self.wavelengths = numpy.zeros(self.nelem)

        if fluxes is not None:
            self.fluxes = numpy.asarray(fluxes)
        else:
            self.fluxes = numpy.zeros(self.nelem)

        if fluxerrs is not None:
            self.fluxerrs = numpy.asarray(fluxerrs)
        else:
            self.fluxerrs = numpy.zeros(self.nelem)

        if dqs is not None:
            self.dqs = numpy.asarray(dqs)
        else:
            self.dqs = numpy.zeros(self.nelem)
#--------------------



"""
.. module:: cosspectrum
   :synopsis: Class definitions for a COS Spectrum object.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from builtins import str
#--------------------
# External Imports
#--------------------
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__

#--------------------

class COSSpectrum():
    """
    Defines a COS spectrum, including wavelegnth, flux, flux errors, and DQ_WGT
    values.  A COS spectrum consists of N segments (N = {2,3}) stored as a dict
    object.  Each of these dicts contain a COSSegment object that contains the
    wavelengths, fluxes, flux errors, etc.

    :raises: ValueError
    """
    def __init__(self, optical_element, band=None, cos_segments=None,
                 orig_file=None):
        """
        Create a COSSpectrum object given a band choice (must be "FUV" or
        "NUV").

        :param optical_element: The string representation of the optical
            element used for this observation, e.g., "G140L".

        :type optical_element: str

        :param band: Which band is this spectrum for ("FUV" or "NUV")?

        :type band: str

        :param cos_segments: [Optional] COSSegment objects to populate the
            COSSpectrum with.

        :type cos_segments: dict

        :param orig_file: Original FITS file read to create the spectrum
            (includes full path).

        :type orig_file: str

        :raises: ValueError
        """

        # Record the optical element.
        self.optical_element = optical_element

        # Record the original file name.
        self.orig_file = orig_file

        if band.strip().upper() == "FUV":
            # Record the band name.
            self.band = band

            if cos_segments is not None:
                if len(cos_segments) == 2:
                    self.segments = {'FUVA':cos_segments['FUVA'],
                                     'FUVB':cos_segments['FUVB']}
                elif len(cos_segments) == 1 and 'FUVA' in cos_segments:
                    self.segments = {'FUVA':cos_segments['FUVA']}
                elif len(cos_segments) == 1 and 'FUVB' in cos_segments:
                    self.segments = {'FUVB':cos_segments['FUVB']}
                else:
                    raise ValueError("Band is specified as " +
                                     band.strip().upper() +
                                     ", expected 1 or 2 COSSegment objects as"
                                     " a list but received " +
                                     str(len(cos_segments))+".")
            else:
                # Otherwise `cos_segments` was not supplied, so create with
                # an empty `segments` property.
                self.segments = {'FUVA':COSSegment(), 'FUVB':COSSegment()}

        elif band.strip().upper() == "NUV":
            # Record the band name.
            self.band = band

            if cos_segments is not None:
                if len(cos_segments) == 3:
                    self.segments = {'NUVA':cos_segments['NUVA'],
                                     'NUVB':cos_segments['NUVB'],
                                     'NUVC':cos_segments['NUVC']}
                elif len(cos_segments) == 2:
                    if 'NUVA' in cos_segments and 'NUVB' in cos_segments:
                        self.segments = {'NUVA':cos_segments['NUVA'],
                                         'NUVB':cos_segments['NUVB']}
                    if 'NUVA' in cos_segments and 'NUVC' in cos_segments:
                        self.segments = {'NUVA':cos_segments['NUVA'],
                                         'NUVC':cos_segments['NUVC']}
                    if 'NUVB' in cos_segments and 'NUVC' in cos_segments:
                        self.segments = {'NUVB':cos_segments['NUVB'],
                                         'NUVC':cos_segments['NUVC']}
                elif len(cos_segments) == 1:
                    if 'NUVA' in cos_segments:
                        self.segments = {'NUVA':cos_segments['NUVA']}
                    if 'NUVB' in cos_segments:
                        self.segments = {'NUVB':cos_segments['NUVB']}
                    if 'NUVC' in cos_segments:
                        self.segments = {'NUVC':cos_segments['NUVC']}
                else:
                    raise ValueError("Band is specified as " +
                                     band.strip().upper() +
                                     ", expected 1, 2, or 3 COSSegment objects"
                                     " as a list but received " +
                                     str(len(cos_segments))+".")
            else:
                # Otherwise `cos_segments` was not supplied, so create with
                # an empty `segments` property.
                self.segments = {'NUVA':COSSegment(), 'NUVB':COSSegment(),
                                 'NUVC':COSSegment()}

        else:
            raise ValueError("Must specify band=\"FUV\" or band=\"NUV\".")
#--------------------

#--------------------
class COSSegment():
    """
    Defines a spectrum from a COS segment.  The data (wavelength, flux, flux
    errors) are stored as numpy ndarrays.  A scalar int property provides the
    number of elements in this segment.
    """
    def __init__(self, nelem=None, wavelengths=None, fluxes=None, fluxerrs=None,
                 dqs=None):
        """
        Create a COSSegment object, default to empty values.  Allows user to
        preallocate space if they desire by setting "nelem" but not providing
        lists/arrays on input right away.

        :param nelem: Number of elements for this segment's spectrum.

        :type nelem: int

        :param wavelengths: [Optional] List of wavelengths in this segment's
            spectrum.

        :type wavelengths: list

        :param fluxes: [Optional] List of fluxes in this segment's spectrum.

        :type fluxes: list

        :param fluxerrs: [Optional] List of flux uncertainties in this segment's
            spectrum.

        :type fluxerrs: list

        :param dqs: [Optional] List of Data Quality (DQ) flags in this segment's
            spectrum.

        :type dqs: list
        """

        # <DEVEL> Should it be required to have `nelem` > 0 *OR* specify
        # arrays on input?  Otherwise they are pre-allocated to empty lists.
        # </DEVEL>
        if nelem is not None:
            self.nelem = nelem
        else:
            self.nelem = 0

        if wavelengths is not None:
            self.wavelengths = numpy.asarray(wavelengths)
        else:
            self.wavelengths = numpy.zeros(self.nelem)

        if fluxes is not None:
            self.fluxes = numpy.asarray(fluxes)
        else:
            self.fluxes = numpy.zeros(self.nelem)

        if fluxerrs is not None:
            self.fluxerrs = numpy.asarray(fluxerrs)
        else:
            self.fluxerrs = numpy.zeros(self.nelem)

        if dqs is not None:
            self.dqs = numpy.asarray(dqs)
        else:
            self.dqs = numpy.zeros(self.nelem)
#--------------------



"""
.. module:: haspspectrum
   :synopsis: Defines the class for HASP spectra.

.. moduleauthor:: Rob Swaters <rswaters@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
#--------------------
# External Imports
#--------------------
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__

#--------------------

class HASPSpectrum():
    """
    Defines a HASP 1D spectrum, including wavelegnth, flux, and flux errors.

    :raises: ValueError
    """

    def __init__(self, nelem=None, wavelengths=None, fluxes=None, fluxerrs=None,
                 dqs=None, plot_info=None, orig_file=None):
        """
        Create a HASPSpectrum object, default to empty values.  Allows user
        to preallocate space if they desire by setting "nelem" but not providing
        lists/arrays on input right away.

        :param nelem: Number of elements for this segment's spectrum.

        :type nelem: int

        :param wavelengths: List of wavelengths in this segment's spectrum.

        :type wavelengths: list

        :param fluxes: List of fluxes in this segment's spectrum.

        :type fluxes: list

        :param fluxerrs: List of flux uncertainties in this segment's spectrum.

        :type fluxerrs: list

        :param dqs: List of data quality flags.

        :type dqs: list

        :param plot_info: dictionary with information for the preview plot.

        :type plot_info: dict

        :param orig_file: Original FITS file read to create the spectrum
            (includes full path).

        :type orig_file: str
        """

        # Record the original file name along with the list of associations.
        self.orig_file = orig_file

        # <DEVEL> Should it be required to have `nelem` > 0 *OR* specify
        # arrays on input?  Otherwise they are pre-allocated to empty
        # lists. </DEVEL>
        if nelem is not None:
            self.nelem = nelem
        else:
            self.nelem = 0

        if wavelengths is not None:
            self.wavelengths = numpy.asarray(wavelengths)
        else:
            self.wavelengths = numpy.zeros(self.nelem)

        if fluxes is not None:
            self.fluxes = numpy.asarray(fluxes)
        else:
            self.fluxes = numpy.zeros(self.nelem)

        if fluxerrs is not None:
            self.fluxerrs = numpy.asarray(fluxerrs)
        else:
            self.fluxerrs = numpy.zeros(self.nelem)

        if dqs is not None:
            self.dqs = numpy.asarray(dqs)
        else:
            self.dqs = numpy.zeros(self.nelem)

        if plot_info is not None:
            self.plot_info = plot_info
        else:
            self.plot_info = {}

#--------------------



"""
.. module:: test_placeholder
   :synopsis: Placeholder unit test until actual tests are implemented.

.. moduleauthor:: Scott W. Fleming
"""

import pytest

def test_placeholder():
    assert 1 == 1



"""
.. module:: stitch_components
   :synopsis: Given a COS or STIS spectrum, will stitch each segment/order,
       respectively, into a contiguous array.  Does not handle any overlap in
       wavelength between sections.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from __future__ import division
import os
import sys
from builtins import range
from builtins import str
#--------------------
# External Imports
#--------------------
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots.utils.specutils.is_bad_dq import is_bad_dq
from spec_plots.utils.specutils.get_flux_stats import get_flux_stats
from spec_plots.utils.specutils.edge_trim import edge_trim
from spec_plots.utils.specutils_cos.cosspectrum import COSSpectrum
from spec_plots.utils.specutils_stis.stis1dspectrum import STISExposureSpectrum
from spec_plots import __version__
#--------------------

#--------------------
def stitch_components(input_exposure, n_consecutive, flux_scale_factor,
                      fluxerr_scale_factor, segment_names=None):
    """
    Given a COSSpectrum or STISExposureSpectrum object, stitches each
    segment/order, respectively, into a contiguous array.

    :param input_exposure: The COS segment or STIS exposure spectrum to stitch.

    :type input_exposure: COSSpectrum or STISExposureSpectrum

    :param n_consecutive: How many consecutive points must pass the test for the
        index to count as the valid start/end of the spectrum?

    :type n_consecutive: int

    :param flux_scale_factor: Max. allowed ratio between the flux and a median
        flux value, used in edge trimming.

    :type flux_scale_factor: float

    :param fluxerr_scale_factor: Max. allowed ratio between the flux uncertainty
        and a median flux uncertainty value, used in edge trimming.

    :type fluxerr_scale_factor: float

    :param segment_names: List of segment names if input_exposure is a
        COSSpectrum object.

    :type segment_names: list

    :returns: dict{numpy array, numpy array, numpy array, str} -- The stitched
        wavelengths, fluxes, flux errors, and an informational plot title in the
        event that all the fluxes had the DQ flag set.  These are packaged in a
        dict.

    :raises: ValueError
    """

    # These lits will contain the stitched spectrum.
    all_wls = []
    all_fls = []
    all_flerrs = []
    all_dqs = []

    # Determine how many pieces there are to stitch, and how to loop through
    # them (different depending on instrument type).
    if isinstance(input_exposure, COSSpectrum):
        if segment_names is not None:
            n_components = len(segment_names)
            loop_iterable = segment_names
            inst_type = "cos"
        else:
            raise ValueError("Must provide a list of segment names for COS"
                             " spectra.")

    elif isinstance(input_exposure, STISExposureSpectrum):
        n_components = len(input_exposure.orders)
        loop_iterable = list(range(n_components))
        inst_type = "stis"

    else:
        raise ValueError("Input must be either a COSSpectrum or"
                         " STISExposureSpectrum object.")

    # Create an array that will keep track whether each component to stitch
    # is filled with bad DQ flags.  The return title will be set to a non-empty
    # string if every component to stitch is filled with bad DQ flags.  Note
    # that for COS the DQ flags are the DQ_WGT bits from the header.
    all_dq_flags = numpy.zeros(n_components)
    return_title = ""

    # Get the wavelengths, fluxes, flux uncertainties, and DQ flags for this
    # component.
    for j_j, j in enumerate(loop_iterable):
        if inst_type == "stis":
            these_wls = input_exposure.orders[j].wavelengths
            these_fls = input_exposure.orders[j].fluxes
            these_flerrs = input_exposure.orders[j].fluxerrs
            these_dqs = input_exposure.orders[j].dqs
            where_bad_dq = numpy.where(is_bad_dq(inst_type, these_dqs))[0]
        elif inst_type == "cos":
            these_wls = input_exposure.segments[j].wavelengths
            these_fls = input_exposure.segments[j].fluxes
            these_flerrs = input_exposure.segments[j].fluxerrs
            these_dqs = input_exposure.segments[j].dqs
            where_bad_dq = numpy.where(is_bad_dq(inst_type, these_dqs))[0]
        n_elems = len(these_wls)

        # Calculate some statistics for this component.
        median_flux, median_fluxerr, fluxerr_95th = get_flux_stats(these_fls,
                                                                   these_flerrs)

        # Find the start and end indices using the edge trimming function.
        (start_index_nodq, end_index_nodq, start_index_withdq,
         end_index_withdq) = edge_trim(inst_type, these_fls, these_flerrs,
                                       these_dqs, n_consecutive, median_flux,
                                       flux_scale_factor, median_fluxerr,
                                       fluxerr_scale_factor, fluxerr_95th)

        # Check if this component has all bad DQ flags.
        if len(where_bad_dq) == n_elems:
            all_dq_flags[j_j] = 1

        # Only append the parts of this order's spectrum that pass the edge
        # trimming, but do NOT trim too much.
        if (start_index_withdq+1) / float(n_elems) <= 0.1:
            start_index_to_use = start_index_withdq
        elif (start_index_nodq+1) / float(n_elems) <= 0.1:
            start_index_to_use = start_index_nodq
        else:
            start_index_to_use = 0

        if abs(end_index_withdq) / float(n_elems) <= 0.1:
            end_index_to_use = end_index_withdq
        elif abs(end_index_nodq) / float(n_elems) <= 0.1:
            end_index_to_use = end_index_nodq
        else:
            end_index_to_use = -1

        # Append the portion of this component to the spectrum we are
        # building up.
        if end_index_to_use == -1:
            all_wls += list(these_wls[start_index_to_use:])
            all_fls += list(these_fls[start_index_to_use:])
            all_flerrs += list(these_flerrs[start_index_to_use:])
            all_dqs += list(these_dqs[start_index_to_use:])
        else:
            all_wls += list(these_wls[start_index_to_use:end_index_to_use+1])
            all_fls += list(these_fls[start_index_to_use:end_index_to_use+1])
            all_flerrs += list(these_flerrs[start_index_to_use:end_index_to_use+
                                            1])
            all_dqs += list(these_dqs[start_index_to_use:end_index_to_use+1])

    # If every single order had all DQ flags, then we want to print a warning
    # on the plot. """
    if sum(all_dq_flags) == n_components:
        if inst_type == 'stis':
            return_title = "Warning: All fluxes have bad DQ."
        elif inst_type == 'cos':
            return_title = "Warning: All fluxes have bad DQ_WGT."

    # Convert these to numpy arrays.
    # <DEVEL> Should these just be numpy arrays to begin with? </DEVEL>
    all_wls = numpy.asarray(all_wls)
    all_fls = numpy.asarray(all_fls)
    all_flerrs = numpy.asarray(all_flerrs)
    all_dqs = numpy.asarray(all_dqs)

    # Make sure the spectrum is sorted in wavelength.
    # <DEVEL> Not sure how to handle components that might overlap in
    # wavelength space, but are not monotonically increasing/decreasing as each
    # component is stitched. </DEVEL>
    sorted_indexes = numpy.argsort(all_wls)
    all_wls = all_wls[sorted_indexes]
    all_fls = all_fls[sorted_indexes]
    all_flerrs = all_flerrs[sorted_indexes]
    all_dqs = all_dqs[sorted_indexes]

    return {"wls":all_wls, "fls":all_fls, "flerrs":all_flerrs, "dqs":all_dqs,
            "title":return_title}
#--------------------



"""
.. module:: readspec
   :synopsis: Reads in a JWST spectrum from a FITS file.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from __future__ import print_function
import sys
#--------------------
# External Imports
#--------------------
from astropy.io import fits
from astropy.table import Table
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__
from spec_plots.utils.specutils_jwst.jwstspectrum import JWSTSpectrum

#--------------------

#--------------------
def readspec(input_file):
    """
    Reads in a JWST spectrum FITS file (x1d, x1dints) and
    returns the wavelengths, fluxes, flux uncertainties, and DQ values.

    :param input_file: Name of input FITS file.

    :type input_file: str

    :returns: JWSTSpectrum -- The spectroscopic data (wavelength, flux,
        flux error, etc):

    :raises: KeyError
    """

    with fits.open(input_file) as hdulist:
        # Read the data from the "EXTRACT1D" FITS extension.
        try:
            jwst_table = Table.read(hdulist["EXTRACT1D"],
                                        unit_parse_strict='silent')
        except KeyError:
            sys.exit("*** MAKE_JWST_SPEC_PREVIEWS ERROR: EXTRACT1D extension"
                         " not found in FITS file.")

        # For time series observations (TSO), many exposures are included in
        # a single EXTRACT1D. In this case, a preview plot is only made for
        # a single exposure (the first one). Remake a simplified Table that
        # contains the columns of interest as columns rather than lists in a
        # row entry.
        extract_one_row = False
        if len(jwst_table["WAVELENGTH"].shape) == 2:
            extract_one_row = True

        # Extract wavelength, fluxes, flux uncertainties, and DQ flags for
        # each segment.
        try:
            if not extract_one_row:
                wavelength_col = jwst_table["WAVELENGTH"].data
            else:
                wavelength_col = jwst_table[0]["WAVELENGTH"]
            wavelength_unit = jwst_table["WAVELENGTH"].unit
        except KeyError:
            sys.exit("*** MAKE_JWST_SPEC_PREVIEWS ERROR: WAVELENGTH column not"
                  " found in first extension's binary table.")

        try:
            if not extract_one_row:
                flux_col = jwst_table["FLUX"].data
            else:
                flux_col = jwst_table[0]["FLUX"]
            flux_unit = jwst_table["FLUX"].unit
        except KeyError:
            sys.exit("*** MAKE_JWST_SPEC_PREVIEWS ERROR: FLUX column not"
                         " found in first extension's binary table.")

        try:
            if not extract_one_row:
                fluxerr_col = jwst_table["ERROR"].data
            else:
                fluxerr_col = jwst_table[0]["ERROR"]
        except KeyError:
            try:
                if not extract_one_row:
                    fluxerr_col = jwst_table["FLUX_ERROR"].data
                else:
                    fluxerr_col = jwst_table[0]["FLUX_ERROR"]
            except KeyError:
                sys.exit("*** MAKE_JWST_SPEC_PREVIEWS ERROR: neither ERROR "
                          "nor FLUX_ERROR column found in first "
                          "extension's binary table.")

        try:
            if not extract_one_row:
                dq_col = jwst_table["DQ"].data
            else:
                dq_col = jwst_table[0]["DQ"]
        except KeyError:
            sys.exit("*** MAKE_JWST_SPEC_PREVIEWS ERROR: DQ column not found"
                  " in first extension's binary table.")

        # Create JWSTSpectrum object.
        return_spec = JWSTSpectrum(wavelength_col, flux_col,
                                       fluxerr_col, dq_col,
                                       wavelength_unit, flux_unit,
                                       orig_file=input_file)

        return return_spec
#--------------------



"""
.. module:: readspec
   :synopsis: Reads in a COS spectrum from a FITS file.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from __future__ import print_function
import sys
#--------------------
# External Imports
#--------------------
from astropy.io import fits
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__
from spec_plots.utils.specutils_cos.check_segments import check_segments
from spec_plots.utils.specutils_cos.cosspectrum import COSSpectrum, COSSegment

#--------------------

#--------------------
def readspec(input_file):
    """
    Reads in a COS spectrum FITS file (x1d, x1dsum, or x1dsum{1,2,3,4}) and
    returns the wavelengths, fluxes, flux uncertainties, and DQ_WGT values for
    the two (FUV segments) or three (NUV stripes).

    :param input_file: Name of input FITS file.

    :type input_file: str

    :returns: COSSpectrum -- The spectroscopic data (wavelength, flux,
        flux error, etc):

    :raises: KeyError
    """

    with fits.open(input_file) as hdulist:

        # Read the data from the first extension.  For COS, the spectra are
        # always stored as tables in the first FITS extension.
        cos_tabledata = hdulist[1].data

        # Extract the SEGMENTS.  This is either an (up-to) 2-element array
        # of ["FUVA", "FUVB"], or an (up-to) 3-element array of ["NUVA", "NUVB",
        # "NUVC"].
        try:
            segment_arr = cos_tabledata.field("SEGMENT")
        except KeyError:
            sys.exit("*** MAKE_HST_SPEC_PREVIEWS ERROR: SEGMENT column not"
                         " found in first extension's binary table.")

        # Determine which band this is (NUV, FUV).
        band = check_segments(segment_arr, input_file)

        # Extract the optical element from the primary header.
        try:
            optical_element = hdulist[0].header["OPT_ELEM"]
        except KeyError:
            sys.exit("*** MAKE_HST_SPEC_PREVIEWS ERROR: OPT_ELEM keyword not"
                    " found in the primary header.")

        # Extract the number of elements (n_wavelengths, n_fluxes, etc.) for
        # each segment.  The dimension will match the array of segment names.
        try:
            nelems_arr = cos_tabledata.field("NELEM")
        except KeyError:
            sys.exit("*** MAKE_HST_SPEC_PREVIEWS ERROR: NELEM column not found"
                  " in first extension's binary table.")

        # Extract wavelength, fluxes, flux uncertainties, and DQ flags for
        # each segment.  These will be mxn tables, where m is the number of
        # segment names, and n is the number of elements from the nelems array.
        try:
            wavelength_table = cos_tabledata.field("WAVELENGTH")
        except KeyError:
            sys.exit("*** MAKE_HST_SPEC_PREVIEWS ERROR: WAVELENGTH column not"
                  " found in first extension's binary table.")

        try:
            flux_table = cos_tabledata.field("FLUX")
        except KeyError:
            sys.exit("*** MAKE_HST_SPEC_PREVIEWS ERROR: FLUX column not"
                         " found in first extension's binary table.")

        try:
            fluxerr_table = cos_tabledata.field("ERROR")
        except KeyError:
            sys.exit("*** MAKE_HST_SPEC_PREVIEWS ERROR: ERROR column not found"
                  " in first extension's binary table.")

        try:
            dq_table = cos_tabledata.field("DQ_WGT")
        except KeyError:
            sys.exit("*** MAKE_HST_SPEC_PREVIEWS ERROR: DQ_WGT column not found"
                  " in first extension's binary table.")

        # Create COSSegment objects to populate the COSSpectrum object with.
        if band == 'FUV':
            # Try creating FUVA COSSegment object.
            try:
                fuva_index = numpy.where(segment_arr == "FUVA")[0][0]
            except IndexError:
                # Then there is no FUVA segment, normally this happends if
                # there is only one segment present (in which case it's the
                # other one of the two).
                fuva_index = None
            if fuva_index is not None:
                fuva_cossegment = COSSegment(nelem=nelems_arr[fuva_index],
                                             wavelengths=
                                             wavelength_table[fuva_index, :],
                                             fluxes=flux_table[fuva_index, :],
                                             fluxerrs=fluxerr_table[
                                                 fuva_index, :],
                                             dqs=dq_table[fuva_index, :])

            # Try creating FUVB COSSegment object.
            try:
                fuvb_index = numpy.where(segment_arr == "FUVB")[0][0]
            except IndexError:
                # Then there is no FUVB segment, normally this happens if
                # there is only one segment present (in which case it's the
                # other one of the two).
                fuvb_index = None
            if fuvb_index is not None:
                fuvb_cossegment = COSSegment(nelem=nelems_arr[fuvb_index],
                                             wavelengths=wavelength_table[
                                                 fuvb_index, :],
                                             fluxes=flux_table[fuvb_index, :],
                                             fluxerrs=fluxerr_table[
                                                 fuvb_index, :],
                                             dqs=dq_table[fuvb_index, :])
        elif band == 'NUV':
            # Try creating NUVA COSSegment object.
            try:
                nuva_index = numpy.where(segment_arr == "NUVA")[0][0]
            except IndexError:
                # Then there is no NUVA segment, not sure this is possible
                # but handle the case anyways.
                nuva_index = None
            if nuva_index is not None:
                nuva_cossegment = COSSegment(nelem=nelems_arr[nuva_index],
                                             wavelengths=wavelength_table[
                                                 nuva_index, :],
                                             fluxes=flux_table[nuva_index, :],
                                             fluxerrs=fluxerr_table[
                                                 nuva_index, :],
                                             dqs=dq_table[nuva_index, :])

            # Try creating NUVB COSSegment object.
            try:
                nuvb_index = numpy.where(segment_arr == "NUVB")[0][0]
            except IndexError:
                # Then there is no NUVB segment, not sure this is possible
                # but handle the case anyways.
                nuvb_index = None
            if nuvb_index is not None:
                nuvb_cossegment = COSSegment(nelem=nelems_arr[nuvb_index],
                                             wavelengths=wavelength_table[
                                                 nuvb_index, :],
                                             fluxes=flux_table[nuvb_index, :],
                                             fluxerrs=fluxerr_table[
                                                 nuvb_index, :],
                                             dqs=dq_table[nuvb_index, :])

            # Try creating NUVC COSSegment object.
            try:
                nuvc_index = numpy.where(segment_arr == "NUVC")[0][0]
            except IndexError:
                # Then there is no NUVC segment, not sure this is possible
                # but handle the case anyways.
                nuvc_index = None
            if nuvc_index is not None:
                nuvc_cossegment = COSSegment(nelem=nelems_arr[nuvc_index],
                                             wavelengths=wavelength_table[
                                                 nuvc_index, :],
                                             fluxes=flux_table[nuvc_index, :],
                                             fluxerrs=fluxerr_table[
                                                 nuvc_index, :],
                                             dqs=dq_table[nuvc_index, :])

        # Create COSSpectrum object.
        if band == 'FUV':
            # Handle case where both are supplied.
            if fuva_index is not None and fuvb_index is not None:
                return_spec = COSSpectrum(optical_element, band=band,
                                          cos_segments={'FUVA':fuva_cossegment,
                                                        'FUVB':fuvb_cossegment},
                                          orig_file=input_file)
            elif fuva_index is not None:
                # Handle cases where only one is supplied.
                return_spec = COSSpectrum(optical_element, band=band,
                                          cos_segments={'FUVA':fuva_cossegment},
                                          orig_file=input_file)
            elif fuvb_index is not None:
                return_spec = COSSpectrum(optical_element, band=band,
                                          cos_segments={'FUVB':fuvb_cossegment},
                                          orig_file=input_file)
            else:
                raise ValueError("Neither FUVA or FUVB segments were found,"
                                 " unable to create COS spectrum object.")

        elif band == 'NUV':
            # Handle case where all three are supplied.
            if (nuva_index is not None and nuvb_index is not None and
                    nuvc_index is not None):
                return_spec = COSSpectrum(optical_element, band=band,
                                          cos_segments={'NUVA':nuva_cossegment,
                                                        'NUVB':nuvb_cossegment,
                                                        'NUVC':nuvc_cossegment},
                                          orig_file=input_file)
            elif nuva_index is not None and nuvb_index is not None:
                # Handle cases where only two are supplied.
                return_spec = COSSpectrum(optical_element, band=band,
                                          cos_segments={'NUVA':nuva_cossegment,
                                                        'NUVB':nuvb_cossegment},
                                          orig_file=input_file)
            elif nuva_index is not None and nuvc_index is not None:
                return_spec = COSSpectrum(optical_element, band=band,
                                          cos_segments={'NUVA':nuva_cossegment,
                                                        'NUVC':nuvc_cossegment},
                                          orig_file=input_file)
            elif nuvb_index is not None and nuvc_index is not None:
                return_spec = COSSpectrum(optical_element, band=band,
                                          cos_segments={'NUVB':nuvb_cossegment,
                                                        'NUVC':nuvc_cossegment},
                                          orig_file=input_file)
            elif nuva_index is not None:
                # Handle cases where only one is supplied.
                return_spec = COSSpectrum(optical_element, band=band,
                                          cos_segments={'NUVA':nuva_cossegment},
                                          orig_file=input_file)
            elif nuvb_index is not None:
                return_spec = COSSpectrum(optical_element, band=band,
                                          cos_segments={'NUVB':nuvb_cossegment},
                                          orig_file=input_file)
            elif nuvc_index is not None:
                return_spec = COSSpectrum(optical_element, band=band,
                                          cos_segments={'NUVC':nuvc_cossegment},
                                          orig_file=input_file)
            else:
                raise ValueError("None of the NUVA, NUVB, or NUVC segments"
                                 " were found, unable to create COS spectrum"
                                 " object.")

        return return_spec
#--------------------



"""
.. module:: rms
   :synopsis: Calculates the root-mean-square of an array of input values.
       Optionally, a constant offset value can be subtracted before calculating
       the rms.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
import math
#--------------------
# External Imports
#--------------------
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__

#--------------------

def rms(values, offset=0.):
    """
    Calculates the RMS about some offset (default offset is 0.)

    :param values: Array of values to compute the rms of.

    :type values: numpy.ndarray

    :param offset: Optional offset to compute the rms about.  Defaults to 0.0.

    :type offset: float

    :returns: float -- A scalar float containing the rms about the offset.
    """
    return math.sqrt(numpy.nanmean([(x-offset)**2 for x in values]))
#--------------------
